# -*- coding: utf-8 -*-
import logging

import click

from trailblazer.add import add_cmd
from trailblazer import utils

log = logging.getLogger(__name__)


@click.command()
@click.argument('cust_dirs', type=click.Path(exists=True), nargs=-1)
@click.pass_context
def scan(context, cust_dirs):
    """Scan directory(ies) for analyses."""
    for cust_dir in cust_dirs:
        log.debug("scanning customer dir: %s", cust_dir)
        sampleinfo_files = utils.scan(cust_dir)
        for sampleinfo in sampleinfo_files:
            log.debug("adding analysis: %s", sampleinfo)
            with open(sampleinfo, 'r') as stream:
                context.invoke(add_cmd, qcsampleinfo=stream)



# -*- coding: utf-8 -*-
import os

from flask import abort, Flask, render_template, request, redirect
from flask_alchy import Alchy
from flask_bootstrap import Bootstrap
import sqlalchemy as sqa

from trailblazer.store import Analysis, Model

app = Flask(__name__)
DEBUG = False
SECRET_KEY = os.environ.get('TRAILBLAZER_SECRET_KEY') or 'thisIsNotSecret!'
BOOTSTRAP_SERVE_LOCAL = True
SQLALCHEMY_DATABASE_URI = os.environ['SQLALCHEMY_DATABASE_URI']
if 'mysql' in SQLALCHEMY_DATABASE_URI:  # pragma: no cover
    SQLALCHEMY_POOL_RECYCLE = 3600
SQLALCHEMY_TRACK_MODIFICATIONS = False
TEMPLATES_AUTO_RELOAD = True
app.config.from_object(__name__)

Bootstrap(app)
db = Alchy(app, Model=Model)


@app.route('/')
def index():
    recent_query = (Analysis.query.filter_by(status='completed')
                                  .order_by(Analysis.started_at.desc())
                                  .limit(10))
    fail_query = (Analysis.query.filter_by(status='failed')
                                .order_by(Analysis.started_at.desc())
                                .limit(20))
    running_query = (Analysis.query.filter_by(status='running')
                                   .order_by(Analysis.started_at.desc()))
    return render_template('index.html', fails=fail_query,
                           runnings=running_query, recents=recent_query)


@app.route('/analyses')
def analyses():
    """Show all analyses."""
    page_num = int(request.args.get('page', 1))
    query_str = request.args.get('query_str')
    query = Analysis.query.order_by(Analysis.started_at.desc())
    if query_str:
        query = query.filter(sqa.or_(Analysis.case_id.contains(query_str),
                                     Analysis.status == query_str))
    page = query.paginate(page=page_num, per_page=30)
    return render_template('analyses.html', analyses=page, query_str=query_str)


@app.route('/analyses/<case_id>')
def analysis(case_id):
    """Show history for an analysis."""
    analyses = (Analysis.query.filter_by(case_id=case_id)
                              .order_by(Analysis.started_at.desc()))
    return render_template('analysis.html', analyses=analyses, case_id=case_id)


@app.route('/comments/<analysis_id>', methods=['POST'])
def comments(analysis_id):
    """Interact with comments."""
    analysis_obj = Analysis.query.get(analysis_id)
    if analysis_obj is None:
        return abort(404)
    new_comment = request.form['comment']
    analysis_obj.comment = new_comment
    db.commit()
    return redirect(request.referrer)



# -*- coding: utf-8 -*-
import logging

from alchy import Manager

from .models import Analysis, Model

log = logging.getLogger(__name__)


def connect(uri):
    log.debug('open connection to database: %s', uri)
    manager = Manager(config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=Model)
    return manager


def analyses(analysis_id=None, since=None, is_ready=False):
    """List added analyses."""
    query = Analysis.query.order_by(Analysis.started_at.desc())

    if since:
        log.debug("filter analyses on date: %s", since)
        query = query.filter(Analysis.started_at > since)

    if analysis_id:
        log.debug("filter analyses on id pattern: %s", analysis_id)
        query = query.filter(Analysis.case_id.contains(analysis_id))

    if is_ready:
        query = query.filter_by(status='completed', is_deleted=False)

    return query



# -*- coding: utf-8 -*-
import logging

import click

from trailblazer.store import api

log = logging.getLogger(__name__)


@click.command()
@click.option('-d', '--database')
@click.option('-r', '--reset', is_flag=True)
@click.pass_context
def init(context, database, reset):
    """Setup the analysis package."""
    db_uri = database or context.obj['database']
    log.info("setup a new database: %s", db_uri)
    manager = api.connect(db_uri)
    if reset:
        manager.drop_all()
    manager.create_all()



# -*- coding: utf-8 -*-
import codecs
import logging
import os
import pkg_resources

import click
import yaml

from .log import init_log
from trailblazer.store import api

log = logging.getLogger(__name__)


class EntryPointsCLI(click.MultiCommand):

    """Add subcommands dynamically to a CLI via entry points."""

    def _iter_commands(self):
        """Iterate over all subcommands as defined by the entry point."""
        return {entry_point.name: entry_point for entry_point in
                pkg_resources.iter_entry_points('trailblazer.subcommands.1')}

    def list_commands(self, ctx):
        """List the available commands."""
        commands = self._iter_commands()
        return commands.keys()

    def get_command(self, ctx, name):
        """Load one of the available commands."""
        commands = self._iter_commands()
        if name not in commands:
            click.echo("no such command: {}".format(name))
            ctx.abort()
        return commands[name].load()


def build_cli(title, Model):
    """Build base cli from scratch."""
    version = pkg_resources.get_distribution(title).version

    @click.group(cls=EntryPointsCLI)
    @click.option('-c', '--config', default="~/.{}.yaml".format(title),
                  type=click.Path(), help='path to config file')
    @click.option('-d', '--database', help='path/URI of the SQL database')
    @click.option('-l', '--log-level', default='INFO')
    @click.option('--log-file', type=click.Path())
    @click.version_option(version, prog_name=title)
    @click.pass_context
    def root(context, config, database, log_level, log_file):
        """Interact with CLI."""
        init_log(logging.getLogger(), loglevel=log_level, filename=log_file)
        log.info("{}: version {}".format(title, version))

        # read in config file if it exists
        if os.path.exists(config):
            with codecs.open(config) as conf_handle:
                context.obj = yaml.load(conf_handle)
        else:
            context.obj = {}

        if database:
            context.obj['database'] = database
        if context.obj.get('database'):
            context.obj['manager'] = api.connect(context.obj['database'])

    return root



# -*- coding: utf-8 -*-
from datetime import date
import logging

import click
from path import path

from . import api

log = logging.getLogger(__name__)


@click.command()
@click.argument('case_id')
@click.pass_context
def delete(context, case_id):
    """Delete an analysis and files."""
    manager = context.obj['manager']
    analysis_obj = api.analyses(analysis_id=case_id).one()
    if analysis_obj.is_deleted:
        click.echo("this analysis is already deleted")
    else:
        click.echo("you are about to delete: {}".format(analysis_obj.root_dir))
        if click.confirm('are you sure?'):
            path(analysis_obj.root_dir).rmtree_p()
            analysis_obj.is_deleted = True
            manager.commit()
            click.echo("removed: {}".format(analysis_obj.root_dir))


@click.command('list')
@click.option('-p', '--pretty', is_flag=True)
@click.option('-l', '--limit', default=10)
@click.option('-s', '--since', nargs=3, type=int)
@click.option('-c', '--config', is_flag=True)
@click.argument('analysis_id', required=False)
@click.pass_context
def list_cmd(context, pretty, limit, since, config, analysis_id):
    """List added analyses."""
    if since:
        since = date(*since)
    query = api.analyses(analysis_id=analysis_id, since=since, is_ready=config)

    if query.first() is None:
        log.warn('sorry, no analyses found')
    else:
        if config:
            paths = (analysis.config_path for analysis in query)
            click.echo(' '.join(paths))
        else:
            for analysis in query.limit(limit):
                click.echo(analysis.to_json(pretty=pretty))



# -*- coding: utf-8 -*-
from datetime import datetime

from path import path


def scan(root_dir):
    """Scan root for MIP analyses.

    Will look for qc sample info files.
    """
    files = path(root_dir).glob('**/**/**/*_sampleInfo.yaml')
    return files


def get_sacctout(analysis_root):
    """Return the path to the sacct STDOUT file."""
    files = path(analysis_root).glob('bwa/info/Sacct_*.stdout.txt')
    if len(files) > 1:
        # pick the "latest" file - increments with number in the filename
        return files[-1]
    try:
        return files[0]
    except IndexError:
        # no file found - expect the analysis already is running
        return None


def inspect_error(sacct_stream):
    """Check STDOUT of the sacct command for failiures."""
    rows = (line.split() for line in sacct_stream)
    # filter rows that begin with a SLURM job id
    relevant_rows = (row for row in rows if row[0].isdigit())
    # filter rows that have a FAILED status
    nonsuccess_rows = (row for row in relevant_rows if row[9] not in
                       ('COMPLETED', 'RUNNING'))
    # format the results
    nonsuccess_jobs = [convert_job(row) for row in nonsuccess_rows]
    return nonsuccess_jobs


def convert_job(row):
    """Convert sacct row to dict."""
    return {
        'id': row[0],
        'name': row[1],
        'time': row[6],
        'start': datetime.strptime(row[7], '%Y-%m-%dT%H:%M:%S'),
        'state': row[9]
    }



# -*- coding: utf-8 -*-
from datetime import datetime
import logging

import click
from path import path
import yaml

from trailblazer.store import Analysis
from trailblazer import utils
from trailblazer.exc import MissingFileError, UnknownError

log = logging.getLogger(__name__)


def parse_sampleinfo(sampleinfo):
    """Extract basic info from loaded sampleinfo file.

    Args:
        sampleinfo (dict): loaded contents from sampleinfo file
    """
    fam_key = sampleinfo.keys()[0]
    fam_data = sampleinfo[fam_key][fam_key]
    sample_ids = sampleinfo[fam_key].keys()
    sample_ids.remove(fam_key)
    analysis_start = fam_data['AnalysisDate']
    analysis_type = fam_data['AnalysisType']
    analysis_out = path(fam_data['Program']['QCCollect']['OutDirectory'])
    customer = fam_data['InstanceTag'][0]
    case_id = "{}-{}".format(customer, fam_key)
    cluster_const = analysis_out.parent.parent.parent
    config_path = cluster_const.joinpath(analysis_type, fam_key,
                                         "{}_config.yaml".format(fam_key))
    values = {
        'case_id': case_id,
        'pipeline': 'mip',
        'pipeline_version': fam_data['MIPVersion'],
        'type': analysis_type,
        'root_dir': analysis_out.parent.parent,
        'started_at': analysis_start,
        'config_path': config_path,
        'samples': sample_ids,
        'analysis_status': fam_data['AnalysisRunStatus'],
    }
    return values


def parse_status(analysis_status, analysis_start, sacct_out=None):
    """Parse the status for a MIP analysis."""
    if analysis_status in ('Finished', 'Archived', 'Archiving'):
        return dict(status='completed')
    else:
        # analysis status is "notFinished" - find out why!
        long_since_start = (datetime.now() - analysis_start).seconds > 86400
        if sacct_out is None:
            # we can't really tell if something went wrong
            if long_since_start:
                # if the analysis should've finished but hasn't
                return dict(status='failed', failed_step='time')
            else:
                # assume the analysis is still running without fails
                return dict(status='running')
        else:
            error_jobs = utils.inspect_error(sacct_out)
            if len(error_jobs) > 0:
                # we found some failed jobs!
                first_fail = error_jobs[0]
                if analysis_start > first_fail['start']:
                    # the analysis has been restarted
                    # the Sacct errors don't belong to this analysis!
                    if long_since_start:
                        # long since start, still not completed
                        return dict(status='failed', failed_step='time')
                    else:
                        return dict(status='running')
                else:
                    # Sacct output belongs to this analysis
                    return dict(status='failed',
                                failed_step=first_fail['name'],
                                failed_at=first_fail['start'])
            else:
                # the status in QC sample info should be "Finished"...
                raise UnknownError()


def build_analysis(sampleinfo):
    """Build database record from analysis output."""
    values = parse_sampleinfo(sampleinfo)
    sacct_out = utils.get_sacctout(values['root_dir'])
    args = dict(analysis_status=values['analysis_status'],
                analysis_start=values['started_at'])
    if sacct_out:
        with open(sacct_out) as stream:
            status = parse_status(sacct_out=stream, **args)
    else:
        status = parse_status(**args)

    values.update(status)
    del values['analysis_status']
    new_analysis = Analysis(**values)
    new_analysis.samples = values['samples']
    return new_analysis


def is_newest(data):
    """Check if the analysis is up to date."""
    fam_key = data.keys()[0]
    version = data[fam_key][fam_key].get('MIPVersion')
    return version is not None and version.startswith('v3.')


def is_updated(old_analysis, new_analysis):
    """Check if a new analysis has been updated."""
    return any((old_analysis.status != new_analysis.status,
                old_analysis.failed_step != new_analysis.failed_step,
                old_analysis.failed_at != new_analysis.failed_at))


@click.command('add')
@click.argument('qcsampleinfo', type=click.File('r'))
@click.pass_context
def add_cmd(context, qcsampleinfo):
    """Add an analysis to the database."""
    manager = context.obj['manager']
    data = yaml.load(qcsampleinfo)
    family_id = data.keys()[0]
    if is_newest(data):
        try:
            new_analysis = build_analysis(data)
            filters = dict(case_id=new_analysis.case_id,
                           started_at=new_analysis.started_at)
            old_analysis = (Analysis.query.filter_by(**filters)
                                          .order_by(Analysis.started_at.desc())
                                          .first())
            if old_analysis and not is_updated(old_analysis, new_analysis):
                log.debug("nothing updated since last analysis: %s",
                          new_analysis.case_id)
            else:
                if old_analysis and old_analysis.status == 'running':
                    # replace if status was 'running'
                    log.debug("deleting existing analysis: %s",
                              new_analysis.case_id)
                    old_analysis.delete()
                    manager.commit()
                manager.add_commit(new_analysis)
                log.info("added new analysis: %s - %s", new_analysis.case_id,
                         new_analysis.status)
        except MissingFileError:
            log.error("missing status file for: %s", family_id)
    else:
        log.debug("analysis version not supported: %s", family_id)



# -*- coding: utf-8 -*-
from datetime import datetime
import json

import alchy
from sqlalchemy import Column, types


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial
    raise TypeError('Type not serializable')


class JsonModel(alchy.ModelBase):

    def to_json(self, pretty=False):
        """Serialize Model to JSON."""
        kwargs = dict(indent=4, sort_keys=True) if pretty else dict()
        return json.dumps(self.to_dict(), default=json_serial, **kwargs)


Model = alchy.make_declarative_base(Base=JsonModel)


class Analysis(Model):

    """Analysis record."""

    id = Column(types.Integer, primary_key=True)
    case_id = Column(types.String(128))

    # metadata
    pipeline = Column(types.Enum('mip'))
    pipeline_version = Column(types.String(32))
    started_at = Column(types.DateTime)
    completed_at = Column(types.DateTime)
    status = Column(types.Enum('running', 'completed', 'failed'))
    root_dir = Column(types.Text)
    config_path = Column(types.Text)
    type = Column(types.Enum('exomes', 'genomes'))
    failed_step = Column(types.String(128))
    failed_at = Column(types.DateTime)
    comment = Column(types.Text)
    is_deleted = Column(types.Boolean, default=False)
    _samples = Column(types.Text)

    @property
    def samples(self):
        return self._samples.split(',')

    @samples.setter
    def samples(self, sample_list):
        """Serialize a list of sample ids."""
        self._samples = ','.join(sample_list)
